#ifndef TYPETAGGER_BHMM_HPP
#define TYPETAGGER_BHMM_HPP

#include <vector>
#include <algorithm> // unary_function
#include <iostream>
#include <boost/foreach.hpp>
#include <boost/program_options.hpp>
#include <boost/random/uniform_01.hpp>
#include <boost/math/distributions/gamma.hpp>
#include <boost/random/gamma_distribution.hpp>

#include "type_config.hpp"
#include "lexicon_state.hpp"
#include "scorer.hpp"

namespace typetagger {

struct bhmm_gibbs_options {
  size_type iterations;
  size_type num_states;
  int resample;
  float_type alpha_emission;   // token
  float_type alpha_transition; // token
  float_type alpha_tag;        // type
  float_type alpha_feats;      // type

  bool approx;
  bool use_dict;
  int init;

  std::ostream& display(std::ostream& outs) const
  {
    outs << "iterations " << iterations << std::endl
         << "num_states " << num_states << std::endl
         << "resample " << resample << std::endl
         << "alpha_emission " << alpha_emission << std::endl
         << "alpha_transition " << alpha_transition << std::endl
         << "alpha_tag " << alpha_tag << std::endl
         << "alpha_feats " << alpha_feats << std::endl
         << "approx " << approx << std::endl
         << "use_dict " << use_dict << std::endl
         << "init " << init << std::endl
         << "resample " << resample << std::endl;
      ;
    return outs;
  }

  bhmm_gibbs_options(const boost::program_options::variables_map& vm)
    :
    iterations( vm["iterations"].as<size_type>() ),
    num_states( vm["num-states"].as<size_type>() ),
    resample( vm.count("resample") ? vm["resample"].as<int>() : 0),
    alpha_emission( vm["alpha-emission"].as<float_type>() ),
    alpha_transition( vm["alpha-transition"].as<float_type>() ),
    alpha_tag( vm.count("alpha-tag") ? vm["alpha-tag"].as<float_type>() : -9 ),
    alpha_feats( vm.count("alpha-features") ? vm["alpha-features"].as<float_type>() : -9 ),
    approx( vm.count("approx") ),
    use_dict( vm.count("use-dict") ),
    init ( vm.count("init") ? vm["init"].as<int>() : 0 )
  {}
};

std::ostream& operator<<(std::ostream& outs, const bhmm_gibbs_options& opts)
{
  return opts.display(outs);
}

class bhmm {
  static const double gamma_shape;
  static const double gamma_scale;

  const std::vector<size_type> wseq_;
  const std::vector<size_type> tseq_;
  const boost::multi_array<size_type,2> wfeats_;
  const std::vector<std::string> wlex_;
  const std::vector<std::string> tlex_;
  const typetagger::scorer scorer_;
  const std::vector<std::vector<size_type> > dict_;

  size_type sample_new_tag(const lexicon_state& state, 
                           random_uniform_real_type& rand01,
                           const size_type w, 
                           const size_type oldt
                           ) const
  {
    const std::vector<float_type> logprobs = state.logprobs (w, oldt);

    assert ( logprobs.size() == 1 + state.num_states() );
    assert ( logprobs[0] == -lexicon_state::INF ); // first bin impossible
    const boost::tuple<std::size_t,float_type,float_type> sample = typetagger::sample_from_logs( rand01, logprobs );
    const std::size_t index = sample.get<0>();
    return index;
  }

  boost::tuple<float_type,
               std::vector<size_type>,
               float_type,
               std::vector<size_type>,
               float_type,float_type,float_type,
               float_type>
  score(const lexicon_state& state) const
  {
    return scorer_.score( state.tagging_dict() );
  }

  void print_state_score( std::ostream& outs, const int iter, const lexicon_state& state ) const
  {
    const boost::tuple<float_type,
      std::vector<size_type>,
      float_type,
      std::vector<size_type>,
      float_type,float_type,float_type,
      float_type> scores = score( state );
    outs << "iter " << iter
         << " greedy1to1 " << scores.get<0>()
         << " many1 " << scores.get<2>() 
         << " vm " << scores.get<4>() //<< ' ' << scores.get<5>() << ' ' << scores.get<6>()
         << " type " << scores.get<7>()
      //<< " lllh_slow " << state.lllh_slow()
         << " lllh " << state.lllh()
         << std::endl;
  }

  void run_gibbs_one_iter(const std::size_t iter, 
                          random_uniform_real_type& rand01,
                          lexicon_state& state, 
                          const std::vector<size_type>& vocab
                          ) const
  {
    //BOOST_FOREACH( const size_type w, state.vocab() )
    BOOST_FOREACH( const size_type w, vocab )
      {
        assert ( std::fabs(state.lllh() - state.lllh_slow()) < 1e-5 );
        const size_type old_tag = state.remove_word_type(w);
        //assert ( state.validate() );
        assert ( std::fabs(state.lllh() - state.lllh_slow()) < 1e-5 );
        const size_type new_tag = sample_new_tag(state, rand01, w, old_tag);
        assert ( new_tag >= 1 );
        assert ( new_tag <= state.num_states() );
        state.add_word_type(w, new_tag);
        //assert ( state.validate() );
      }
      assert ( state.validate() );
      assert ( std::fabs(state.lllh() - state.lllh_slow()) < 1e-5 );
  }

  boost::tuple<double,double>
  evaluate_logfunction(const lexicon_state& state,
                       const std::vector<double>& hyperparameters,
                       const std::size_t i) const
  {
    static const boost::math::gamma_distribution<> gamma( bhmm::gamma_shape, bhmm::gamma_scale );
    const double lllh = state.lllh(hyperparameters);
    const double prior = std:: log( boost::math::pdf(gamma, hyperparameters[i]) );
    //std::cout << "llllh = " << lllh << " prior " << prior << std::endl;
    return boost::make_tuple(lllh, prior);
  }
                           

  void resample_hyperparameters (const int it, 
                                 random_uniform_real_type& rand01,
                                 boost::variate_generator<boost_rng_type&, boost::uniform_int<> >& randint, 
                                 lexicon_state& state) const
  {
    assert ( rand01.min() == 0 );
    assert ( rand01.max() == 1 );
    using namespace std;

    const size_t num_samples = 10;
    const vector<double> steps ( 4, 1.0 );
    vector<double> hyperparameters (4);
    hyperparameters [0] = state. alpha_emission();
    hyperparameters [1] = state. alpha_transition();
    hyperparameters [2] = state. alpha_tag();
    hyperparameters [3] = state. alpha_features();

    std::cout << "resampling current values = ";
    std::copy(hyperparameters.begin(), hyperparameters.end(), std::ostream_iterator<double>(std::cout, " "));
    std::cout << std::endl;

    vector<size_t> indices(hyperparameters.size());
    __gnu_cxx::iota (indices.begin(), indices.end(), 0);
    random_shuffle(indices.begin(), indices.end(), randint);

    // slice sample one variable at a time
    // NOTE: we want to sample from prob p(data|hyper)p(hyper) but compute log probs
    //for(size_t i = 0; i < hyperparameters.size(); ++i)
    BOOST_FOREACH(const size_t i, indices) 
      {
        if (hyperparameters[i] < 0) continue;
        for(size_t s = 1; s <= num_samples; ++s)
          {
            // bracket
            const double r01 = rand01();
            const boost::tuple<double,double> values0 = evaluate_logfunction (state, hyperparameters, i); // logprob
            const double l0 = values0.get<0>();
            const double p0 = values0.get<1>();
            const double f0 = l0 + p0;
            const double y0 = f0 + log(r01);
            const double x0 = hyperparameters[i];
            double left = hyperparameters[i] - rand01() * steps[i];
            double right = left + steps[i];
            int trials1 = 0;
            while(left >= 0 && ++trials1)
              {
                hyperparameters[i] = left;
                const boost::tuple<double,double> values1 = evaluate_logfunction (state, hyperparameters, i);
                const double f1 = values1.get<0>() + values1.get<1>();
                if (y0 > f1) break;
                else left -= steps[i];
              }
            left = (left >= 0) ? left : 0;
            while(++trials1 <= 100)
              {
                hyperparameters[i] = right;
                const boost::tuple<double,double> values1 = evaluate_logfunction (state, hyperparameters, i);
                const double f1 = values1.get<0>() + values1.get<1>();
                //std::cout << "right = " << right << " " << y0 << " " << f1 << std::endl;
                if (y0 > f1) break;
                else right += steps[i];
              }
            
            // sample next state
            assert ( left < right );
            //std::cout << "x0 " << x0 << " f0 " << f0 << " r " << r01 << " y0 " << y0 << " [ " << left << " " << right << " ]" << std::endl;
            int trials2 = 0;
            double ff1 = -1;
            double ll1 = -1;
            double pp1 = -1;
            while(++trials2)
              {
                const double x1 = left + (right - left) * rand01();
                hyperparameters[i] = x1;
                const boost::tuple<double,double> values1 = evaluate_logfunction (state, hyperparameters, i);
                ll1 = values1.get<0>();
                pp1 = values1.get<1>();
                const double f1 = ll1 + pp1;
                ff1 = f1;
                //std::cout << "x1 " << x1 << " f1 " << f1 << std::endl;
                if (f1 > y0) break;
                else
                  {
                    if (x1 > x0) right = x1;
                    else left = x1;
                  }
              }
            if (s == num_samples)
              {
                std::cout << "resampling var " << i << " sample " << s << " " 
                          << " r01 " << r01
                          << " x " << x0 << " -> " << hyperparameters[i]
                          << " f " << f0 << " -> " << ff1
                          << " l " << l0 << " -> " << ll1
                          << " evals " << trials1 << " " << trials2 << std::endl;
              }
            if (i == 0)      state. alpha_emission ( hyperparameters[i] );
            else if (i == 1) state. alpha_transition ( hyperparameters[i] );
            else if (i == 2) state. alpha_tag ( hyperparameters[i] );
            else if (i == 3) state. alpha_features ( hyperparameters[i] );
          }
      }
    std::cout << "resampled new values = ";
    std::copy(hyperparameters.begin(), hyperparameters.end(), std::ostream_iterator<double>(std::cout, " "));
    std::cout << std::endl;
  }

public:
  bhmm(const char* const wseq_filename,
       const char* const tseq_filename,
       const char* const wlex_filename,
       const char* const tlex_filename,
       const char* const wfeats_filename,
       const char* const dict_filename)
    : wseq_( slurp_into_padded_stream<size_type>(wseq_filename, 0) ),
      tseq_( slurp_into_padded_stream<size_type>(tseq_filename, 0) ),
      wfeats_( read_type_features<size_type>(wfeats_filename) ),
      wlex_( read_lexicon(wlex_filename) ),
      tlex_( read_lexicon(tlex_filename) ),
      scorer_( typetagger::scorer(wseq_,tseq_) ),
      dict_( read_tagging_dict(dict_filename, wlex_.size()) )
  {
    assert(wseq_.size() == tseq_.size());

    //typedef boost::multi_array_types::index_range range;
    //const boost::multi_array<size_type,2>::array_view<2>::type v = wfeats_ [ boost::indices[0][range()] ];
    //for(std::size_t i = 0; i < wfeats_.shape()[1]; ++i) std::cout << wfeats_[0][i] << ' ';
    //std::cout << std::endl;
    //for(std::size_t i = 0; i < wfeats_.shape()[1]; ++i) std::cout << wfeats_[wfeats_.shape()[0]-1][i] << ' ';
    //std::cout << std::endl;

    //std::copy( v.begin(), v.end(), std::ostream_iterator<size_type>(std::cout," ")); std::cout << std::endl;
    //std::copy(wfeats_[0].begin(), wfeats_[0].end(), std::ostream_iterator<size_type>(std::cout," ")); std::cout << std::endl;
    //std::copy(wfeats_[1].begin(), wfeats_[1].end(), std::ostream_iterator<size_type>(std::cout," ")); std::cout << std::endl;
    //std::copy(wfeats_.back().begin(), wfeats_.back().end(), std::ostream_iterator<size_type>(std::cout," ")); std::cout << std::endl;

    if (wseq_.size() != tseq_.size()) throw;
  }

  lexicon_state run_gibbs(const bhmm_gibbs_options& opts, boost_rng_type& rng) const
  {
    using namespace std;
    if (opts.use_dict && dict_.empty()) throw;
    if (opts.alpha_feats >= 0 && wfeats_.num_elements() == 0) { std::cout << "error: empty features!" << endl; throw; }

    boost::random::gamma_distribution<> gamma_dist( bhmm::gamma_shape, bhmm::gamma_scale );
    boost::variate_generator<boost_rng_type&, boost::random::gamma_distribution<> > gamrand(rng, gamma_dist);

    const double alpha_emission = opts.alpha_emission > 0 && opts.resample < 0 ? gamrand() : opts.alpha_emission ;
    const double alpha_transition = opts.alpha_transition > 0 && opts.resample < 0 ? gamrand() : opts.alpha_transition;
    const double alpha_tag = opts.alpha_tag > 0 && opts.resample < 0 ? gamrand() : opts.alpha_tag;
    const double alpha_feats = opts.alpha_feats > 0 && opts.resample < 0 ? gamrand() : opts.alpha_feats;

    lexicon_state state(rng,
                        wseq_, tseq_, wfeats_, wlex_, tlex_, dict_, 
                        opts.num_states, 
                        alpha_emission,
                        alpha_transition,
                        alpha_tag,
                        alpha_feats,
                        opts.approx,
                        opts.use_dict,
                        opts.init);

    boost::uniform_real<> uni_dist01(0,1);
    random_uniform_real_type rand01(rng, uni_dist01);
    boost::uniform_int<> uni_dist_int;
    boost::variate_generator<boost_rng_type&, boost::uniform_int<> > randint(rng, uni_dist_int);

    assert ( state.validate() );
    print_state_score( std::cout, -1, state );
    vector<size_type> vocab(state.vocab().begin(), state.vocab().end());
    const size_t resample_frequency = opts.resample > 0 ? opts.resample : -opts.resample;
    assert ( std::fabs(state.lllh() - state.lllh_slow()) < 1e-5 );
    for(size_t it = 1; it <= opts.iterations; ++it)
      {
        random_shuffle(vocab.begin(), vocab.end(), randint);
        run_gibbs_one_iter(it, rand01, state, vocab);
        print_state_score( std::cout, it, state );
        if (opts.resample != 0 
            && it != opts.iterations
            && it % resample_frequency == 0)
          {
            resample_hyperparameters (it, rand01, randint, state);
          }
      }
    return state;
  }
}; // class bhmm
const double bhmm::gamma_shape = 10.0;
const double bhmm::gamma_scale = 0.1;

} // namespace

#endif // TYPETAGGER_BHMM_HPP
