#ifndef TYPETAGGER_DATAIO_HPP
#define TYPETAGGER_DATAIO_HPP

#include <vector>
#include <fstream>
#include <sstream>
#include <boost/scoped_ptr.hpp>
#include <boost/multi_array.hpp>
#include "utilities.hpp"

namespace typetagger {

std::vector<std::string> read_lexicon(const char* const filename)
{
  const std::vector<std::string> lines = typetagger::slurp(filename, true);
  std::vector<std::string> r;
  r.push_back(""); // one-based
  r.insert( r.end(), lines.begin(), lines.end() );
  return r;
}


template<typename T>
std::vector<T> slurp_into_padded_stream(const char* const filename, const T pad)
{
  using namespace std;

  const vector<string> lines = slurp(filename, true);
  size_t nwords = 0;
  vector<vector<string> > sents;
  BOOST_FOREACH(const string& line, lines)
    {
      sents.push_back( vector<string>() );
      vector<string>& tokens = sents.back();
      boost::algorithm::split(tokens, line, boost::is_any_of(" \t"), boost::token_compress_on);
      nwords += tokens.size();
    }

  vector<T> v;
  v.reserve(nwords + lines.size() + 1);
  v.push_back( pad );
  BOOST_FOREACH(const vector<string>& sent, sents)
    {
      BOOST_FOREACH(const string& token, sent)
        {
          v.push_back( boost::lexical_cast<T>(token) );
        }
      v.push_back(pad);
    }
  assert(v.size() == nwords + lines.size() + 1);
  return v;
}

template<typename T>
void dump_tagging_dict(const char* const outfilename, const std::vector<T>& v)
{
  assert ( v[0] == 0 );
  boost::scoped_ptr<std::ofstream> outfile ( safe_ofopen(outfilename) );
  for(size_t w = 1; w < v.size(); ++w)
    {
      *outfile << w << ' ' << v[w] << std::endl;
    }
  outfile->close();
}

std::vector<std::vector<size_type> > read_tagging_dict(const char* const filename, const std::size_t num_words)
{
  // make dict filename optional by returning empty vector if not provided
  if (filename == 0) return std::vector<std::vector<size_type> >();

  std::vector<std::vector<size_type> > v(num_words);
  
  boost::scoped_ptr<std::ifstream> infile ( safe_ifopen(filename) );
  std::string line;
  std::istringstream iss;
  std::size_t w = 1;
  while(getline(*infile, line) )
    {
      iss.str(line);
      size_type t;
      while(iss >> t)
        {
          std::cout << "w = " << w << " t = " << t << std::endl;
          v[w]. push_back (t);
        }
      iss.clear();
      ++w;
    }
  infile->close();
  assert (w <= num_words + 1);
  return v;
}

template<typename T>
boost::multi_array<T,2> read_type_features(const char* filename)
{
  using namespace std;

  // make feature file optional by returning empty array
  if (filename == 0) return boost::multi_array<T,2> ( boost::extents [0][0] );

  //cout << "read_type_features: opening " << filename << endl;
  const vector<string> lines = typetagger::slurp(filename, true);
  //cout << "read_type_features: opened " << filename << endl;

  if (lines.empty()) 
    return boost::multi_array<T,2>( boost::extents [lines.size()] [0] );

  vector<string> tokens;
  boost::algorithm::split(tokens, lines[0], boost::is_any_of(" \t"), boost::token_compress_on);
  const size_t dim = tokens.size();
  
  boost::multi_array<T,2> v ( boost::extents [lines.size()] [dim] );

  for(size_t i = 0; i < lines.size(); ++i)
    {
      const string& line = lines[i];
      vector<string> tokens;
      boost::algorithm::split(tokens, line, boost::is_any_of(" \t"), boost::token_compress_on);
      assert ( dim == tokens.size() );
      vector<T> vv;
      for(size_t j = 0; j < tokens.size(); ++j)
        {
          v [i] [j] = boost::lexical_cast<T>(tokens[j]);
        }
    }
  return v;
}

} // namespace 
#endif // TYPETAGGER_DATAIO_HPP
