#ifndef TYPETAGGER_DIRICHLET_MULT_HPP
#define TYPETAGGER_DIRICHLET_MULT_HPP

#include <tr1/unordered_set> // std::tr1::unordered_set
#include <cmath>
#include <vector>
#include <iostream>
#include <algorithm>
#include <boost/foreach.hpp>
#include <boost/noncopyable.hpp>

#include "type_config.hpp"

namespace typetagger {

//class dirichlet_mult : private boost::noncopyable {
class dirichlet_mult { // made alpha_ non-const
public:
  typedef std::size_t key_type;
private:
  typedef std::tr1::unordered_set<size_type> key_set;
  float_type alpha_;
  std::vector<int> counts_;
  int total_;
  key_set keys_;
public:
  dirichlet_mult(const float_type alpha = -1.0, 
                 const size_type num_keys = 0)
    :
    alpha_(alpha),
    counts_( std::vector<int>(num_keys, 0) ),
    total_(0),
    keys_( key_set() )
  {
    assert (num_keys >= 0);
    //assert(alpha >= 0);
    //std::fill_n( counts_.data(), counts_.num_elements(), 0);
  }

  dirichlet_mult(const dirichlet_mult& other)
    :
    alpha_(other.alpha_),
    counts_(other.counts_),
    total_(other.total_),
    keys_(other.keys_)
  {
  }

  // noncopyable prohibits
  dirichlet_mult& operator=(const dirichlet_mult& other)
  {
    if (this != &other)
      {
        this->alpha_ = other.alpha_;
        this->counts_ = other.counts_;
        this->total_ = other.total_;
        this->keys_ = other.keys_;
      }
    return *this;
  }

  bool operator==(const dirichlet_mult& other) const
  {
    assert ( this->alpha_ == other.alpha_ );
    assert ( this->counts_ == other.counts_ );
    assert ( this->total_ == other.total_ );
    assert ( keys_equal(this->keys_,other.keys_) );
    return this->alpha_ == other.alpha_
      && this->counts_ == other.counts_
      && this->total_ == other.total_
      && keys_equal(this->keys_,other.keys_);
  }

  bool operator!=(const dirichlet_mult& other) const
  {
    return ! (*this == other);
  }

  int get_counts(const key_type k) const
  {
    return counts_[k];
  }

  const std::vector<int>& counts() const
  {
    return counts_;
  }

  int total() const
  {
    return total_;
  }

  float_type alpha() const { return alpha_; }
  float_type& alpha() { return alpha_; }

  std::ostream& display(std::ostream& outs) const
  {
    outs << "a = " << alpha_ << " counts = ";
    std::copy(counts_.begin(), counts_.end(), std::ostream_iterator<int>(outs," "));
    return outs;
  }

  static
  bool keys_equal(const std::tr1::unordered_set<size_type>& a,
                  const std::tr1::unordered_set<size_type>& b)
  {
    BOOST_FOREACH(const size_type k, a)
      {
        if (b.find(k) == b.end())
          return false;
      }
    BOOST_FOREACH(const size_type k, b)
      {
        if (a.find(k) == a.end())
          return false;
      }
    return true;
  }

  bool validate() const
  {
    const bool total_match = total_ == std::accumulate(counts_.begin(), counts_.end(), 0);
    const bool total_gez = total_ >= 0;
    const bool counts_gez = std::count_if(counts_.begin(), counts_.end(), std::bind2nd(std::less<int>(), 0)) == 0;
    const bool counts_le_total =  std::count_if(counts_.begin(), counts_.end(), std::bind2nd(std::greater<int>(), total_)) == 0;

    assert ( total_match && total_gez && counts_gez && counts_le_total );
    return total_match && total_gez && counts_gez && counts_le_total;
  }

  int observe(const key_type k, const int count)
  {
    counts_ [k] += count;
    total_ += count;
    const int current_count = counts_ [k];
    assert( current_count >= 0 );
    if ( current_count  > 0 )
      {
        keys_ .insert(k);
      }
    else if ( current_count  == 0 )
      {
        keys_ .erase(k);
      }
    else
      {
        std::cerr << "dirichlet_mult::observe [" << k << "] = "
                  << current_count << " < 0 !!!!  aborting." << std::endl;
        throw;
        abort();
      }
    return current_count;
  }

  bool in_keys(const key_type k) const
  {
    return keys_.find(k) != keys_.end();
  }

  const key_set& keys() const
  {
    return keys_;
  }

  float_type logprob(const key_type k, const size_type vocab_size) const
  {
    assert ( counts_[k] >= 0 );
    assert ( total_ >= 0 );
    assert ( alpha_ >= 0 );
    const float_type numer = alpha_ + counts_[k];
    const float_type denom = alpha_ * vocab_size + total_;
    assert ( numer >= 0 );
    assert ( denom >= 0 );
    assert ( denom >= numer );
    return std::log( numer / denom );
  }
};
std::ostream& operator<<(std::ostream& outs, const dirichlet_mult& a)
{
  return a.display(outs);
}

} // namespace

#endif // TYPETAGGER_DIRICHLET_MULT_HPP
