#ifndef TYPETAGGER_LEXICON_STATE_HPP
#define TYPETAGGER_LEXICON_STATE_HPP

#include <cassert>
#include <vector>
#include <algorithm> // max_element, fill_n, sort
#include <tr1/unordered_set> // std::tr1::unordered_set
#include <tr1/unordered_map> // std::tr1::unordered_set
#include <iostream>
#include <limits>
#include <numeric> // iota
#include <boost/foreach.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/random/uniform_int_distribution.hpp>

#include "type_config.hpp"
#include "dirichlet_mult.hpp"

namespace typetagger {

class lexicon_state {
public:
  static const float_type INF;
private:
  const std::vector<size_type>& wseq_;
  const std::vector<size_type>& tseq_;
  const boost::multi_array<size_type,2>& wfeats_;
  const std::vector<std::string>& wlex_;
  const std::vector<std::string>& tlex_;
  const std::vector<std::vector<size_type> >& dict_;
  const size_type num_states_;
  const std::vector<size_type> all_states_;
  float_type alpha_emission_;     // token
  float_type alpha_transition_;   // token
  float_type alpha_tag_;          // tag
  float_type alpha_feats_;        // tag
  const std::vector<std::vector<size_type> > word_locations_;
  const bool approx_;
  const bool use_dict_;

  const bool use_type_tag_;
  const bool use_type_feats_;

  std::vector<size_type> wt_;
  const std::tr1::unordered_set<size_type> vocab_;

  boost::tuple<std::vector<dirichlet_mult>,
               std::vector<dirichlet_mult> > token_distrs_;
  std::vector<dirichlet_mult>& emission_distr_;
  std::vector<dirichlet_mult>& transition_distr_;

  dirichlet_mult type_tag_distr_;
  const std::vector<std::size_t> max_feat_values_;
  boost::multi_array<dirichlet_mult,2> type_feats_distrs_; // [tag][feat-dim]

  const std::vector<std::size_t> num_feat_values_;

  struct sort_by_freq
  {
    const std::vector<int>& counts;
    sort_by_freq(const std::vector<int>& c) : counts(c) {}
    bool operator()(int a, int b) const { return counts[a] > counts[b]; }
  };
  static
  std::vector<size_type> freq_init_tagging_dict(boost_rng_type& rng_engine,
                                                const std::vector<size_type>& wseq,
                                                const size_type num_states)
  {
    assert( ! wseq.empty() );

    const size_type m = *( std::max_element(wseq.begin(), wseq.end()) );

    // get counts    
    std::vector<int> counts(m+1,0);
    BOOST_FOREACH(const size_type w, wseq)
      {
        if (w > 0) counts [w] += 1;
      }
    
    boost::random::uniform_int_distribution<> dist(1,num_states);
    boost::variate_generator<boost_rng_type&,boost::random::uniform_int_distribution<> > rng(rng_engine, dist);

    // heuristic init: sort by freq, top k types in their own cluster, randomly assign rest
    std::vector<int> indices(m+1);
    __gnu_cxx::iota (indices.begin(), indices.end(), 0);
    std::sort (indices.begin(), indices.end(), sort_by_freq(counts));
    std::vector<size_type> tagdict(m+1, 0);
    for(size_type i = 0; i < indices.size(); ++i)
      {
        const size_type w = indices[i];
        if (counts[w] == 0) break;
        //tagdict [w] = ( i < num_states ) ? i+1 : typetagger::randint(1,num_states);
        tagdict [w] = ( i < num_states ) ? i+1 : rng();
        //std::cout << "w = "<< w << " count " << counts[w] << " state " << tagdict[w] << std::endl;
      }
    return tagdict;
  }  

  static
  std::vector<size_type> random_init_tagging_dict(boost_rng_type& rng_engine,
                                                  const std::vector<size_type>& wseq,
                                                  const size_type num_states)
  {
    assert( ! wseq.empty() );
    const size_type m = *( std::max_element(wseq.begin(), wseq.end()) );
    
    boost::random::uniform_int_distribution<> dist(1,num_states);
    boost::variate_generator<boost_rng_type&,boost::random::uniform_int_distribution<> > rng(rng_engine, dist);

    std::vector<size_type> tagdict(m+1, 0);
    for(size_type w = 1; w < tagdict.size(); ++w)
      {
        //tagdict[w] = typetagger::randint(1,num_states);
        tagdict[w] = rng();
      }
    return tagdict;
  }

  static
  std::tr1::unordered_set<size_type>
  get_vocab(const std::vector<size_type>& wseq)
  {
    // excludes NULL_WORD == 0
    std::tr1::unordered_set<size_type> vocab( wseq.begin(), wseq.end() );
    vocab.erase(0);
    return vocab;
  }

  static
  void observe_type_features ( boost::multi_array<dirichlet_mult,2>& type_feats,
                               const boost::multi_array<size_type,2>& wfeats, 
                               const size_type i, 
                               const size_type tag,
                               const int count )
  {
    assert ( tag != 0 );
    for (size_t j = 0; j < wfeats.shape()[1]; ++j)
      {
        type_feats [ tag ] [ j ]. observe ( wfeats[i][j], count );
      }
  }

  static
  std::vector<std::size_t>
  get_biggest_feature_value ( const boost::multi_array<size_type,2>& wfeats )
  {
    const std::size_t feat_dim = wfeats.shape()[1];
    // find max feature value for each feature dim
    std::vector<std::size_t> biggest (feat_dim, 0);
    for (size_type j = 0; j < wfeats.shape()[1]; ++j)
      {
        biggest[j] = wfeats [0] [j];
        for (size_type i = 1; i < wfeats.shape()[0]; ++i)
          {
            if (wfeats [i][j] > biggest[j])
              biggest[j] = wfeats [i][j];
          }
      }
    return biggest;
  }

  static
  boost::multi_array<dirichlet_mult,2>
  get_type_feature_counts ( const boost::multi_array<size_type,2>& wfeats, 
                            const std::vector<size_type>& wt,
                            const double alpha_feats, 
                            const size_type num_states,
                            const std::vector<std::size_t>& biggest_value )
  {
    const std::size_t feat_dim = wfeats.shape()[1];
    assert ( feat_dim == biggest_value.size() );

    // create 2d dirichlet_mult.  [tag][feat-dim] = pointer
    boost::multi_array<dirichlet_mult,2> type_feats ( boost::extents[num_states+1][feat_dim] );
    for(size_type k = 0; k < num_states+1; ++k)
      {
        for(size_type f = 0; f < feat_dim; ++f)
          {
            if (k == 0)
              type_feats [k][f] = dirichlet_mult ( -1, biggest_value[f]+1 ); // should never be used
            else
              type_feats [k][f] = dirichlet_mult ( alpha_feats, biggest_value[f]+1 );
          }
      }

    // gather counts
    assert ( wt.size() == wfeats.shape()[0]+1 );
    for(size_t i = 0; i < wfeats.shape()[0]; ++i)
      {
        const size_type tag = wt[i+1];
        if ( tag == 0 ) continue; // removed
        observe_type_features ( type_feats, wfeats, i, tag, 1 );
      }
    return type_feats;
  }

  static
  std::vector<std::size_t>
  get_feature_cardinality ( const boost::multi_array<size_type,2>& wfeats )
  {
    std::vector<std::size_t> v ( wfeats.shape()[1] );
    for (std::size_t f = 0; f < wfeats.shape()[1]; ++f)
      {
        std::tr1::unordered_set<size_type> values;
        for (std::size_t i = 0; i < wfeats.shape()[0]; ++i)
          {
            values. insert ( wfeats [i] [f] );
          }
        v [f] = values. size();
      }
    return v;
  }

  static
  boost::tuple<std::vector<dirichlet_mult>,
               std::vector<dirichlet_mult> >
  get_token_counts(const std::vector<size_type>& wseq,
                   const std::vector<size_type>& wt,
                   const size_type num_states,
                   const float_type alpha_emission,
                   const float_type alpha_transition)
  {
    using namespace std;

    const size_type maxw = *( max_element( wseq.begin(), wseq.end() ) );
    assert ( *max_element( wt.begin(), wt.end() ) <= num_states );
    vector<dirichlet_mult> emissions(num_states+1, dirichlet_mult(alpha_emission,maxw+1) );
    vector<dirichlet_mult> transitions(num_states+1, dirichlet_mult(alpha_transition,num_states+1) );
    for(size_t i = 1; i < wseq.size(); ++i)
      {
        const size_type w = wseq[i];
        const size_type t = wt[w];
        const bool removed = (w != 0 && t == 0);
        if ( w != 0 && ! removed )
          {
            assert (t != 0);
            emissions [t].observe(w, 1);
          }
        const size_type w_ = wseq[i-1];
        const size_type t_ = wt[w_];
        const bool removed_ = (w_ != 0 && t_ == 0);
        if ( ! removed && ! removed_ )
          transitions [t_].observe(t, 1);
      }
    return boost::make_tuple(emissions, transitions);
  }

  void observe_type_tag(const size_type t,
                        const int count)
  {
    if (use_type_tag_)
      observe_type_tag(type_tag_distr_, t, count);
  }

  static
  void observe_type_tag(dirichlet_mult& type_tag,
                        const size_type t,
                        const int count)
  {
    type_tag.observe(t, count);
  }

  void observe_type_features(const size_type w,
                             const size_type t,
                             const int count)
  {
    if (use_type_feats_)
      {
        assert ( w > 0 );
        observe_type_features(type_feats_distrs_, wfeats_, w-1, t, count);
      }
  }

  static
  dirichlet_mult
  get_type_tag_counts(const std::vector<size_type>& wt,
                      const size_type num_states,
                      const float_type alpha_tag)
  {
    dirichlet_mult type_tag(alpha_tag, num_states+1);
    BOOST_FOREACH(const size_type t, wt)
      {
        const bool removed = (t == 0);
        if ( ! removed )
          observe_type_tag(type_tag, t, 1);
      }
    return type_tag;
  }

  template<typename T>
  static
  std::vector<T>
  make_vector_seq(const T start, const T end)
  {
    std::vector<T> v;
    for(T i = start; i <= end; ++i)
      v.push_back(i);
    return v;
  }

  static
  std::vector<std::vector<size_type> >
  get_word_locations(const std::vector<size_type>& wseq)
  {
    assert ( ! wseq.empty() );
    const size_type maxw = *( max_element(wseq.begin(), wseq.end()) );
    std::vector<std::vector<size_type> > all_word_locations(maxw+1);
    for(size_type i = 0; i < wseq.size(); ++i)
      {
        const size_type w = wseq[i];
        all_word_locations[w].push_back(i);
      }
    return all_word_locations;
  }


  static
  bool validate_dirichlet_mult_vector(const std::vector<dirichlet_mult>& distrs)
  {
    BOOST_FOREACH(const dirichlet_mult& d, distrs)
      {
        if ( ! d.validate() )
          return false;
      }
    return true;
  }

  static
  float_type logprob_transition2(const dirichlet_mult& d,
                                 const size_type t_,
                                 const size_type t,
                                 const size_type _t,
                                 const size_type num_states)
  {
    float_type numer = 0.0 + d.alpha() + d.get_counts(_t);
    if (t_ == t && t == _t)
      numer += 1.0;
    float_type denom = 0.0 + d.alpha() * num_states + d.total();
    if (t_ == t)
      denom += 1.0;
    assert ( d.get_counts(_t) >= 0 );
    assert ( d.total() >= 0 );
    assert ( numer < denom );
    return std::log( numer / denom );
  }

  std::vector<float_type> logprobs_transition(const size_type w) const
  {
    assert ( wt_ [w] == 0 );
    std::vector<float_type> logprobs(num_states_+1, 0);
    logprobs[0] = -lexicon_state::INF;
    for(size_type t = 1; t <= num_states_; ++t)
      {
        std::vector<dirichlet_mult> distrs(transition_distr_);
        float_type& logprob = logprobs[t];
        BOOST_FOREACH(const size_type i, word_locations_ [w])
          {
            const size_type w_ = wseq_[i-1];
            if (w_ != w)
              {
                const size_type t_ = (w_ == w) ? t : wt_ [w_];
                const size_type num_states = t_ ? num_states_ + 1 : num_states;
                const float_type v1 = distrs[t_].logprob(t, num_states);
                logprob += v1;
                distrs[t_].observe(t, 1);
              }
            const size_type _w = wseq_[i+1];
            const size_type _t = (_w == w) ? t : wt_ [_w];
            const size_type num_states = t ? num_states_ + 1 : num_states;
            const float_type v2 = distrs[t].logprob(_t, num_states);
            logprob += v2;
            distrs[t].observe(_t, 1);
          } 
      }
    return logprobs;
  }

  static
  void increment_transition(std::tr1::unordered_map<size_type,std::tr1::unordered_map<size_type,int> >& counts,
                            const size_type t_,
                            const size_type t)
  {
    if (counts.find(t_) == counts.end())
      counts[t_] = std::tr1::unordered_map<size_type,int>();
    if (counts[t_].find(t) == counts[t_].end())
      counts[t_][t] = 1;
    else
      counts[t_][t] += 1;
  }

  std::vector<float_type> logprobs_transition_fast(const size_type w) const
  {
    assert ( wt_ [w] == 0 );
    //std::vector<float_type> logprobs(num_states_+1, 0);
    //logprobs[0] = -lexicon_state::INF;
    std::vector<float_type> logprobs(num_states_+1, -lexicon_state::INF);
    const bool explore_all_states = !use_dict_ || dict_[w].empty();
    const std::vector<size_type>& state_space = explore_all_states ? all_states_ : dict_[w];
    //for(size_type t = 1; t <= num_states_; ++t)
    BOOST_FOREACH(const size_type t, state_space)
      {
        float_type& logprob = logprobs[t];
        logprob = 0.0;
        using namespace std::tr1;
        unordered_map<size_type,unordered_map<size_type,int> > new_counts;
        BOOST_FOREACH(const size_type i, word_locations_ [w])
          {
            const size_type w_ = wseq_[i-1];
            if (w_ != w)
              {
                const size_type t_ = (w_ == w) ? t : wt_ [w_];
                increment_transition (new_counts, t_, t);
              }
            const size_type _w = wseq_[i+1];
            const size_type _t = (_w == w) ? t : wt_ [_w];
            increment_transition (new_counts, t, _t);
          } 
        typedef unordered_map<size_type,unordered_map<size_type,int> > map2d;
        BOOST_FOREACH(const map2d::value_type& i, new_counts)
          {
            const size_type u = i.first;
            const dirichlet_mult& d = transition_distr_[u];
            const float_type alpha = d.alpha();
            float_type total = d.total();
            const size_type num_states = u ? num_states_ + 1: num_states_;
            typedef unordered_map<size_type,int> map1d;
            BOOST_FOREACH(const map1d::value_type& j, i.second)
              {
                const size_type v = j.first;
                const float_type count = d.get_counts(v);
                const int n = j.second;
                for(int c = 0; c < n; ++c)
                  logprob += std::log( (alpha + count + c) / (alpha * num_states + total + c) );
                total += n;
              }
          }
      }
    return logprobs;
  }

  std::vector<float_type> logprobs_emission(const size_type w) const
  {
    assert ( wt_ [w] == 0 );
    std::vector<float_type> logprobs(num_states_+1, -lexicon_state::INF);
    for(size_type t = 1; t <= num_states_; ++t)
      {
        dirichlet_mult d(emission_distr_[t]);
        assert ( ! d.in_keys( w ) );
        //const size_type vocab_size = vocab_.size();
        const size_type vocab_size = d.keys().size() + 1;
        float_type& logprob = logprobs[t];
        logprob = 0;
        for(std::size_t i = 0; i < word_locations_ [w].size(); ++i)
          {
            logprob += d.logprob( w, vocab_size );
            d.observe(w, 1);
          }
      }
    //std::copy(logprobs.begin(), logprobs.end(), std::ostream_iterator<float_type>(std::cout," ")); std::cout << std::endl;
    return logprobs;
  }

  std::vector<float_type> logprobs_emission_fast(const size_type w) const
  {
    assert ( wt_ [w] == 0 );
    std::vector<float_type> logprobs(num_states_+1, -lexicon_state::INF);
    const bool explore_all_states = !use_dict_ || dict_[w].empty();
    const std::vector<size_type>& state_space = explore_all_states ? all_states_ : dict_[w];
    //for(size_type t = 1; t <= num_states_; ++t)
    BOOST_FOREACH(const size_type t, state_space)
      {
        const dirichlet_mult& d = emission_distr_[t];
        assert ( ! d.in_keys( w ) );
        assert ( d.get_counts(w) == 0 );
        //const size_type vocab_size = vocab_.size();
        const size_type vocab_size = d.keys().size() + 1;
        const float_type total = d.total();
        const float_type alpha = d.alpha();
        const std::size_t n = word_locations_ [w].size();
        // calculate asscending factorial
        float_type& logprob = logprobs[t];
        logprob = 0;
        for(std::size_t i = 0; i < n; ++i)
          logprob += std::log( (alpha + i) / (alpha * vocab_size + total + i) );
      }
    return logprobs;
  }

  std::vector<float_type> logprobs_type_tag_prior(const size_type w) const
  {
    assert ( wt_ [w] == 0 );
    std::vector<float_type> logprobs(num_states_+1, -lexicon_state::INF);
    //logprobs[0] = -lexicon_state::INF;
    const bool explore_all_states = !use_dict_ || dict_[w].empty();
    const std::vector<size_type>& state_space = explore_all_states ? all_states_ : dict_[w];
    //for(size_type t = 1; t <= num_states_; ++t)
    BOOST_FOREACH(const size_type t, state_space)
      {
        logprobs[t] = type_tag_distr_.logprob(t, num_states_);
      }
    return logprobs;
  }

  std::vector<float_type> logprobs_type_feats(const size_type w) const
  {
    assert ( wt_ [w] == 0 );
    std::vector<float_type> logprobs(num_states_+1, -lexicon_state::INF);
    //logprobs[0] = -lexicon_state::INF;
    const bool explore_all_states = !use_dict_ || dict_[w].empty();
    const std::vector<size_type>& state_space = explore_all_states ? all_states_ : dict_[w];
    //for(size_type t = 1; t <= num_states_; ++t)
    BOOST_FOREACH(const size_type t, state_space)
      {
        logprobs[t] = 0;
        for (std::size_t f = 0; f < wfeats_.shape()[1]; ++f)
          logprobs[t] += type_feats_distrs_ [t][f] . logprob ( wfeats_[w-1][f], num_feat_values_[f] );
      }
    return logprobs;
  }


public:
  lexicon_state(boost_rng_type& rng,
                const std::vector<size_type>& wseq,
                const std::vector<size_type>& tseq,
                const boost::multi_array<size_type,2>& wfeats,
                const std::vector<std::string>& wlex,
                const std::vector<std::string>& tlex,
                const std::vector<std::vector<size_type> >& dict,
                const size_type num_states,
                const float_type alpha_emission,
                const float_type alpha_transition,
                const float_type alpha_tag,
                const float_type alpha_feats,
                const bool approx,
                const bool use_dict,
                const int init)
    : wseq_( wseq ),
      tseq_( tseq ),
      wfeats_( wfeats ),
      wlex_( wlex ),
      tlex_( tlex ),
      dict_( dict ),
      num_states_( num_states ),
      all_states_( make_vector_seq<size_type>(1, num_states) ),
      alpha_emission_( alpha_emission ),
      alpha_transition_( alpha_transition ),
      alpha_tag_( alpha_tag ),
      alpha_feats_( alpha_feats ),
      word_locations_ ( get_word_locations(wseq) ),
      approx_ ( approx ),
      use_dict_ ( use_dict ),
    
      use_type_tag_ ( alpha_tag >= 0 ),
      use_type_feats_ ( alpha_feats >= 0 ),

      //wt_( random_init_tagging_dict(wseq, num_states) ),
      wt_( init == 1
           ? random_init_tagging_dict(rng, wseq, num_states)
           : freq_init_tagging_dict(rng, wseq, num_states) ),
      vocab_( get_vocab(wseq) ),

      token_distrs_ ( get_token_counts(wseq, wt_, num_states, alpha_emission, alpha_transition) ),
      emission_distr_ ( token_distrs_.get<0>() ),
      transition_distr_ ( token_distrs_.get<1>() ),

      type_tag_distr_ ( use_type_tag_ 
                        ? get_type_tag_counts(wt_, num_states, alpha_tag) 
                        : dirichlet_mult(0,0) ),
      max_feat_values_ ( use_type_feats_
                         ? get_biggest_feature_value ( wfeats_ )
                         : std::vector<std::size_t>() ),
      type_feats_distrs_ ( use_type_feats_
                           ? get_type_feature_counts ( wfeats_, wt_, alpha_feats, num_states, max_feat_values_) 
                           : boost::multi_array<dirichlet_mult,2> ( boost::extents[0][0] ) ),
      num_feat_values_ ( get_feature_cardinality ( wfeats_ ) )
  {
    assert ( wseq_.size() == tseq_.size() );
    if ( wseq_.size() != tseq_.size() ) throw;
    if ( use_dict_ && dict_.empty() ) throw;
    std::cout << "use_type_tag = " << use_type_tag_ << std::endl;
    std::cout << "use_type_feats = " << use_type_feats_ << std::endl;
    if (init == 1) { std::cout << "initilization = random" << std::endl; } else { std::cout << "initilization = freq" << std::endl; }
    std::cout << "alpha_emission = " << alpha_emission << std::endl;
    std::cout << "alpha_transition = " << alpha_transition << std::endl;
    std::cout << "alpha_tag = " << alpha_tag << std::endl;
    std::cout << "alpha_feats = " << alpha_feats << std::endl;
  }

  const std::tr1::unordered_set<size_type>&
  vocab() const
  { return vocab_; }

  const std::vector<size_type>& tagging_dict() const
  { return wt_; }

  float_type alpha_emission() const { return alpha_emission_; }
  float_type alpha_transition() const { return alpha_transition_; }
  float_type alpha_tag() const { return alpha_tag_; }
  float_type alpha_features() const { return alpha_feats_; }

  float_type alpha_emission(const double a)
  {
    const float_type olda = alpha_emission_;
    alpha_emission_ = a;
    BOOST_FOREACH(dirichlet_mult& d, emission_distr_)
      {
        d.alpha() = a;
      }
    return olda;
  }
  float_type alpha_transition(const double a)
  {
    const float_type olda = alpha_transition_;
    alpha_transition_ = a;
    BOOST_FOREACH(dirichlet_mult& d, transition_distr_)
      {
        d.alpha() = a;
      }
    return olda;
  }
  float_type alpha_tag(const double a)
  {
    const float_type olda = alpha_tag_;
    alpha_tag_ = a;
    type_tag_distr_.alpha() = a;
    return olda;
  }

  float_type alpha_features(const double a)
  {
    float_type olda = alpha_feats_;
    alpha_feats_ = a;
    for(size_type k = 0; k < type_feats_distrs_.shape()[0]; ++k)
      {
        for(size_type f = 0; f < type_feats_distrs_.shape()[1]; ++f)
          {
            if (k > 0)
              {
                assert (type_feats_distrs_[k][f].alpha() == olda);
                type_feats_distrs_[k][f].alpha() = a;
              }
          }
      }

    return olda;
  }

  size_type remove_word_type(const size_type w)
  {
    assert (w > 0);
    assert (wt_ [w] != 0);
    assert (wt_ [0] == 0);

    const size_type t = wt_ [w];
    BOOST_FOREACH(const size_type i, word_locations_ [w])
      {
        assert (i > 0);
        assert( wseq_[i] == w );

        emission_distr_ [t].observe(w, -1);

        const size_type w_ = wseq_ [i-1];
        if (w_ != w)
          {
            const size_type t_ = wt_ [w_];
            transition_distr_ [t_].observe(t, -1);
          }
        const size_type _w = wseq_ [i+1];
        const size_type _t = wt_ [_w];
        transition_distr_ [t].observe(_t, -1);
      }
    observe_type_tag(t,-1);
    observe_type_features(w, t,-1);
    //observe_type_tag(type_tag_distr_, t, -1);
    const size_type old_tag = wt_ [w];
    wt_ [w] = 0;
    assert ( emission_distr_ [old_tag].get_counts(w) == 0 );
    return old_tag;
  }

  void add_word_type(const size_type w, const size_type new_tag)
  {
    assert (w > 0);
    assert (wt_ [w] == 0);
    assert (wt_ [0] == 0);

    wt_ [w] = new_tag;
    observe_type_tag(new_tag, 1);
    observe_type_features(w, new_tag, 1);
    //observe_type_tag(type_tag_distr_, new_tag, 1);
    assert ( emission_distr_ [new_tag].get_counts(w) == 0 );
    BOOST_FOREACH(const size_type i, word_locations_ [w])
      {
        assert (i > 0);
        assert( wseq_[i] == w );
        
        emission_distr_ [new_tag].observe(w, 1);

        const size_type w_ = wseq_ [i-1];
        if (w_ != w)
          {
            const size_type t_ = (w_ != w) ? wt_ [w_] : new_tag;
            transition_distr_ [t_].observe(new_tag, 1);
          }
        const size_type _w = wseq_ [i+1];
        const size_type _t = wt_ [_w];
        transition_distr_ [new_tag].observe(_t, 1);
      }
  }

  bool validate_dirichlet_mult_multi_array(const boost::multi_array<dirichlet_mult,2>& type_feats_distrs) const
  {
    bool feat_ok = true;
    for(std::size_t k = 1; k < type_feats_distrs_.shape()[0]; ++k) 
      {
        for(std::size_t f = 0; f < type_feats_distrs_.shape()[1]; ++f) 
          {
            const bool ok = type_feats_distrs_ [k][f] == type_feats_distrs [k][f];
            assert (ok);
            feat_ok = feat_ok && ok;
          }
      }
    return feat_ok;
  }

  bool validate() const
  {
    const boost::tuple<
            std::vector<dirichlet_mult>,
            std::vector<dirichlet_mult> > 
      token_counts = get_token_counts(wseq_, wt_, num_states_, alpha_emission_, alpha_transition_);


    const bool nullok = wt_[0] == 0;
    const bool to1ok = emission_distr_ == token_distrs_.get<0>();
    const bool to2ok = transition_distr_ == token_distrs_.get<1>();
    const bool emission_ok = validate_dirichlet_mult_vector(emission_distr_);
    const bool transition_ok = validate_dirichlet_mult_vector(transition_distr_);
    
    bool type_ok = true;
    if (use_type_tag_)
      {
        const dirichlet_mult type_tag_distr = get_type_tag_counts(wt_, num_states_, alpha_tag_);
        type_ok = type_ok && type_tag_distr == type_tag_distr_;
      }
    if (use_type_feats_)
      {
        boost::multi_array<dirichlet_mult,2> type_feats_distrs = get_type_feature_counts ( wfeats_, wt_, alpha_feats_, num_states_, max_feat_values_);
        type_ok = type_ok && validate_dirichlet_mult_multi_array(type_feats_distrs);
      }

    assert (nullok && to1ok && to2ok && emission_ok && transition_ok \
            && type_ok );
    
    return nullok && to1ok && to2ok && emission_ok && transition_ok \
      && type_ok;
  }

  size_type num_states() const
  {
    return num_states_;
  }

  std::vector<float_type> logprobs(const size_type w, 
                                   const size_type oldt) const
  {
    if (approx_)
      throw;
    else
      return logprobs_cascade_counts (w, oldt);
  }

  std::vector<float_type> logprobs_cascade_counts (const size_type w, 
                                                   const size_type oldt
                                                   ) const
  {
    using namespace std;
    //const vector<float_type> evals = logprobs_emission(w);
    const vector<float_type> evals = logprobs_emission_fast(w);
    //const vector<float_type> tvals = logprobs_transition(w);
    const vector<float_type> tvals = logprobs_transition_fast(w);

    assert ( evals.size() == num_states_ + 1 );
    assert ( tvals.size() == num_states_ + 1 );

    vector<float_type> logprobs(num_states_+1, 0);
    transform(logprobs.begin(), logprobs.end(), evals.begin(), logprobs.begin(), plus<float_type>());
    transform(logprobs.begin(), logprobs.end(), tvals.begin(), logprobs.begin(), plus<float_type>());
    if (use_type_tag_)
      {
        const vector<float_type> pvals = logprobs_type_tag_prior(w); 
        assert ( pvals.size() == num_states_ + 1 );
        transform(logprobs.begin(), logprobs.end(), pvals.begin(), logprobs.begin(), plus<float_type>());
      }
    if (use_type_feats_)
      {
        const vector<float_type> pvals = logprobs_type_feats(w);
        assert ( pvals.size() == num_states_ + 1 );
        transform(logprobs.begin(), logprobs.end(), pvals.begin(), logprobs.begin(), plus<float_type>());
      }

    assert ( logprobs[0] == -lexicon_state::INF );
    logprobs[0] = -lexicon_state::INF;
    return logprobs;
  }

  std::size_t num_features() const
  {
    return wfeats_.shape()[1];
  }

  double lllh() const
  {
    return lllh ( alpha_emission_,
                  alpha_transition_,
                  alpha_tag_,
                  alpha_feats_);
  }
  
  double lllh(const std::vector<float_type>& alphas) const
  {
    return lllh ( alphas[0],
                  alphas[1],
                  alphas[2],
                  alphas[3] );
  }

  double lllh (const float_type alpha_emission,
               const float_type alpha_transition,
               const float_type alpha_tag,
               const float_type alpha_feats
               ) const
  {
    using namespace std;
    double lllh_tag = 0;
    if (alpha_tag >= 0)
      {
        assert ( type_tag_distr_.counts().size() == num_states_ + 1 );
        int total = 0;
        for(size_t t = 1; t <= num_states_; ++t)
          {
            const int count = type_tag_distr_. get_counts(t);
            for (int i = 0; i < count; ++i)
              {
                lllh_tag += log ( (alpha_tag + i) / (alpha_tag * num_states_ + total + i) );
              }
            total += count;
          }
      }

    double lllh_feats = 0;
    if (alpha_feats >= 0)
      {
        for(size_t f = 0; f < num_features(); ++f)
          {
            for(size_t t = 1; t <= num_states_; ++t)
              {
                assert ( max_feat_values_[f] + 1 == num_feat_values_[f] );
                int total = 0;
                for(size_t v = 0; v <= max_feat_values_[f]; ++v)
                  {
                    const int count = type_feats_distrs_ [t][f]. get_counts (v);
                    for(int i = 0; i < count; ++i)
                      {
                        lllh_feats += log ( (alpha_feats + i) / (alpha_feats * num_feat_values_[f] + total + i) );
                      }
                    total += count;
                  }
              }
          }
      }

    double lllh_tokens = 0;
    if (alpha_emission >= 0)
      {
        //const size_t vocab_size = vocab_.size();
        std::vector<int> totals ( num_states_+1, 0 );
        for(size_t w = 1; w < wt_.size(); ++w)
          {
            const size_t t = wt_[w];
            if (t == 0) continue;
            const int count = emission_distr_ [t]. get_counts(w); 
            const size_t vocab_size = emission_distr_ [t]. keys(). size();
            for(int i = 0; i < count; ++i)
              {
                lllh_tokens += log ( (alpha_emission + i) / (alpha_emission * vocab_size + totals[t] + i) );
              }
            totals [t] += count;
          }
      }
    
    if (alpha_transition >= 0)
      {
        for(size_t t_ = 0; t_ <= num_states_; ++t_)
          {
            const int ns = (t_ == 0) ? num_states_ : num_states_ + 1;
            int total = 0;
            for(size_t t = 0; t <= num_states_; ++t)
              {
                if (t_ == 0 && t == 0) continue;
                const int count = transition_distr_ [t_]. get_counts(t);
                for(int i = 0; i < count; ++i)
                  {
                    lllh_tokens += log ( (alpha_transition + i ) / (alpha_transition * ns + total + i) );
                  }
                total += count;
              }
          }
      }

    return lllh_tag + lllh_feats + lllh_tokens;
  }

  boost::tuple<double,
               boost::multi_array<int,1>,
               boost::multi_array<std::tr1::unordered_map<size_type,int>,2>,
               std::vector<std::tr1::unordered_map<size_type,int> >,
               boost::multi_array<int,2> >
  get_all_counts (const float_type alpha_emission,
                  const float_type alpha_transition,
                  const float_type alpha_tag,
                  const float_type alpha_feats
                  ) const
  {
    // prob (data, latent_vars | hyperparameter), i.e. parameters are integrated out
    using namespace std;

    // types
    double type_lllh = 0;

    boost::multi_array<int,1> type_t ( boost::extents[num_states_+1] );
    std::fill_n (type_t.data(), type_t.num_elements(), 0);
    int total_t = 0;

    if (alpha_tag >= 0)
      {
        BOOST_FOREACH(const size_type t, wt_)
          {
            if (t > 0) 
              {
                type_lllh += log ( (alpha_tag + type_t[t]) / ( alpha_tag * num_states_ + total_t ) );
                type_t [t] += 1;
                ++total_t;
              }
          }
      }

    boost::multi_array<std::tr1::unordered_map<size_type,int>,2> fc (boost::extents[ num_features() ][ num_states_+1 ]); // [feature][tag][value] = count
    boost::multi_array<int,2> fc_total ( boost::extents [ num_features() ][ num_states_+1 ] ); // [feature][tag]
    std::fill_n ( fc_total.data(), fc_total.num_elements(), 0 );
    if (alpha_feats >= 0)
      {
        for(size_t w = 1; w < wt_.size(); ++w)
          {
            const size_type t = wt_[w];
            if (t == 0) continue;
            for (size_t f = 0; f < num_features(); ++f)
              {
                const size_type v = wfeats_[w-1][f];
                assert (num_feat_values_[f] == max_feat_values_[f]+1);
                if (fc [f][t].find(v) == fc[f][t].end())
                  fc [f][t][v] = 0;
                type_lllh += log ( (alpha_feats + fc[f][t][v]) / ( alpha_feats * num_feat_values_[f] + fc_total [f][t]) );
                fc [f][t][v] += 1;
                fc_total [f][t] += 1;
              }
          }
      }

    // tokens
    std::vector<std::tr1::unordered_map<size_type,int> > tok_tw (num_states_+1);
    boost::multi_array<int,1> tok_twsum ( boost::extents[num_states_+1] );
    boost::multi_array<int,2> tok_tt ( boost::extents[num_states_+1][num_states_+1] );
    boost::multi_array<int,1> tok_ttsum ( boost::extents[num_states_+1] );
    std::fill_n ( tok_twsum.data(), tok_twsum.num_elements(), 0);
    std::fill_n ( tok_tt.data(), tok_tt.num_elements(), 0);
    std::fill_n ( tok_ttsum.data(), tok_ttsum.num_elements(), 0);
    double tok_lllh = 0;
    //const size_t vocab_size = vocab_.size(); // approximation. exact: get permissible words for each tag
    std::vector<std::size_t> vocab_sizes ( num_states_ + 1, 0 );
    for(size_t w = 1; w < wt_.size(); ++w)
      {
        const std::size_t tag = wt_[w];
        if ( tag > 0 )  vocab_sizes [ tag ] += 1;
      }
    for(size_t i = 1; i < wseq_.size(); ++i)
      {
        const size_type w = wseq_[i];
        const size_type t = wt_[w];
        const size_type vocab_size = vocab_sizes [t];
        //const size_type vocab_size = vocab_.size(); // approx
        if (tok_tw[t].find(w) == tok_tw[t].end())
          tok_tw[t][w] = 0;
        if (t > 0)
          {
            tok_lllh += log ( ( alpha_emission + tok_tw[t][w] ) / (alpha_emission * vocab_size + tok_twsum[t]) );
          }
        tok_tw [t][w] += 1;
        tok_twsum [t] += 1;


        if (w != 0 && t == 0) continue;
        const size_type w_ = wseq_[i-1];
        const size_type t_ = wt_[w_];
        if (w_ != 0 && t_ == 0) continue;
        const size_type num_states = t_ ? num_states_ + 1: num_states_;
        tok_lllh += log ( (alpha_transition + tok_tt [t_][t] ) / ( alpha_transition * num_states + tok_ttsum[t_] ) );
        tok_tt [t_][t] += 1;
        tok_ttsum [t_] += 1;
      }
    const double lllh = type_lllh + tok_lllh;
    return boost::make_tuple(lllh,
                             type_t,
                             fc,
                             tok_tw,
                             tok_tt);
  }

  double lllh_slow () const
  {
    return lllh_slow ( alpha_emission_,
                       alpha_transition_,
                       alpha_tag_,
                       alpha_feats_);
  }

  double lllh_slow (const float_type alpha_emission,
                    const float_type alpha_transition,
                    const float_type alpha_tag,
                    const float_type alpha_feats
                    ) const
  {
    return get_all_counts(alpha_emission, alpha_transition, alpha_tag, alpha_feats).get<0>();
  }

  

}; // class lexicon_state
const float_type lexicon_state::INF = std::numeric_limits<float_type>::infinity();

} // namespace

#endif // TYPETAGGER_LEXICON_STATE_HPP
