#ifndef TYPETAGGER_SCORER_HPP
#define TYPETAGGER_SCORER_HPP

#include <vector>
#include <algorithm>
#include <tr1/unordered_set> // std::tr1::unordered_set
#include <tr1/unordered_map>
#include <functional>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>

#include "type_config.hpp"

namespace typetagger {
class scorer {
  const std::vector<size_type>& wseq_;
  const std::vector<size_type>& tseq_;
  const size_type num_tokens_;
  const size_type maxt_;
  const std::vector<std::vector<size_type> > gold_tag_locations_;
  const std::vector<size_type> oracle_one_tag_dict_;

  static
  std::vector<size_type> get_oracle_one_tag_dict(const std::vector<size_type>& wseq, const std::vector<size_type>& tseq)
  {
    assert ( wseq.size() == tseq.size() );
    assert ( ! wseq.empty() );
    const size_type max_word = *std::max_element(wseq.begin(), wseq.end());
    std::vector< std::tr1::unordered_map<size_type,int> > counts ( max_word + 1 );
    for(size_type i = 0; i < tseq.size(); ++i)
      {
        const size_type tag = tseq[i];
        const size_type word = wseq[i];
        if (tag == 0) continue;
        if ( counts [word]. find (tag) == counts [word]. end() )
          {
            counts [word] [tag] = 1;
          }
        else
          {
            counts [word] [tag] += 1;
          }
      }
    std::vector<size_type> d ( max_word + 1 );
    for(size_type w = 1; w < counts.size(); ++w)
      {
        int biggest = 0;
        size_type best_tag = 0;
        for(std::tr1::unordered_map<size_type,int>::const_iterator it = counts [w].begin();
            it != counts[w].end();
            ++it)
          {
            if (it->second > biggest)
              {
                biggest = it->second;
                best_tag = it->first;
              }
          }
        d [w] = best_tag;
      }
    return d;
  }

  static
  std::vector<std::vector<size_type> > get_tag_locations(const std::vector<size_type>& tseq, const size_type maxt)
  {
    std::vector<std::vector<size_type> > tlocs(maxt+1);
    for(size_type i = 0; i < tseq.size(); ++i)
      {
        const size_type tag = tseq[i];
        if (tag == 0) continue;
        tlocs[tag].push_back(i);
      }
    return tlocs;
  }
  
  static
  std::vector<std::vector<size_type> >
  get_tag_locations_from_type_dict(const std::vector<size_type>& wseq,
                                   const std::vector<size_type>& dict)
  {
    const size_type maxt = *std::max_element(dict.begin(), dict.end());
    std::vector<std::vector<size_type> > tlocs(maxt+1);
    for(size_type i = 0; i < wseq.size(); ++i)
      {
        const size_type word = wseq[i];
        if (word == 0) continue;
        const size_type tag = dict[word];
        tlocs[tag].push_back(i);
      }
    return tlocs;
  }

  struct state_tag_score
  {
    size_type state;
    size_type tag;
    float_type score;
    state_tag_score(const size_type state,
                    const size_type tag,
                    const float_type score)
      : state(state), tag(tag), score(score)
    {}
  };

  static
  std::vector<state_tag_score> get_mapping_scores(const std::vector<std::vector<size_type> >& state_locs,
                                                  const std::vector<std::vector<size_type> >& tag_locs)
  {
    std::vector<state_tag_score> retval;
    retval.reserve( state_locs.size() * tag_locs.size() );
    for(size_type state = 1; state < state_locs.size(); ++state)
      {
        const std::vector<size_type>& state_loc = state_locs[state]; 
          for(size_type tag = 1; tag < tag_locs.size(); ++tag)
            {
              const std::vector<size_type>& tag_loc = tag_locs[tag];
              std::vector<size_type> matches( std::max( state_loc.size(), tag_loc.size() ) );
              const std::vector<size_type>::const_iterator 
                it = std::set_intersection(state_loc.begin(), state_loc.end(), tag_loc.begin(), tag_loc.end(), matches.begin());
              retval.push_back( state_tag_score( state, tag, static_cast<size_type>(it-matches.begin()) ) );
            }
        }
    return retval;
  }
  
  boost::tuple<float_type, std::vector<state_tag_score> >
  get_greedy1to1_matching(const std::vector<state_tag_score>& sorted_edge_scores, 
                          const size_type maxs) const
  {
    std::tr1::unordered_set<size_type> used_states;
    std::tr1::unordered_set<size_type> used_tags;
    //std::vector<bool> used_states(maxs+1,false);
    //std::vector<bool> used_tags(maxt+1,false);
    float_type correct = 0.0;
    std::vector<state_tag_score> mapping;
    mapping.reserve(maxs);
    BOOST_FOREACH(const state_tag_score& edge, sorted_edge_scores)
      {
        //if ( ! used_states[edge.state] && ! used_tags[edge.tag] )
        if (used_states.find(edge.state) == used_states.end()
            && used_tags.find(edge.tag) == used_tags.end())
          {
            correct += edge.score;
            used_states.insert(edge.state);
            used_tags.insert(edge.tag);
            //used_states[ edge.state ] = true;
            //used_tags[ edge.tag ] = true;
            mapping.push_back(edge);
          }
      }
    const float_type accuracy = correct / num_tokens_;
    return boost::make_tuple(accuracy, mapping);
  }

  boost::tuple<float_type, std::vector<state_tag_score> >
  get_many1_matching(const std::vector<state_tag_score>& sorted_edge_scores, const size_type maxs) const
  {
    std::tr1::unordered_set<size_type> used_states;
    //std::vector<bool> used_states(maxs+1,false);
    float_type correct = 0.0;
    std::vector<state_tag_score> mapping;
    mapping.reserve(maxs);
    BOOST_FOREACH(const state_tag_score& edge, sorted_edge_scores)
      {
        if (used_states.find(edge.state) == used_states.end())
          //if ( ! used_states[ edge.state ] )
          {
            correct += edge.score;
            used_states.insert(edge.state);
            //used_states[ edge.state ] = true;
            mapping.push_back(edge);
          }
      }
    const float_type accuracy = correct / num_tokens_;
    return boost::make_tuple(accuracy, mapping);
  }

  struct state_tag_score_state_less : std::binary_function<bool,state_tag_score,state_tag_score>
  {
    bool operator()(const state_tag_score& a, const state_tag_score& b)
    {
      return a.state < b.state;
    }
  };
  static
  std::vector<size_type> state_tag_score_vect_to_map(const std::vector<state_tag_score>& v)
  {
    assert (! v.empty());

    const size_type maxs = std::max_element(v.begin(), v.end(), state_tag_score_state_less())->state;
    std::vector<size_type> m(maxs+1,0);
    BOOST_FOREACH(const state_tag_score& s, v)
      {
        assert (s.tag > 0);
        m [ s.state ] = s.tag;
      }
    return m;
  }

  static
  boost::tuple<std::vector<double>, 
               std::vector<double>, 
               std::vector< std::tr1::unordered_map<size_type,double> >,
               std::vector< std::tr1::unordered_map<size_type,double> >
               >
  compute_v_measure_stats(const std::vector<state_tag_score>& counts)
  {
    if (counts.empty())
      {
        return boost::make_tuple( std::vector<double>(),
                                  std::vector<double>(),
                                  std::vector< std::tr1::unordered_map<size_type,double> >(),
                                  std::vector< std::tr1::unordered_map<size_type,double> >() );
      }

    size_type max_tag = counts[0].tag;
    size_type max_state = counts[0].state;

    BOOST_FOREACH(const state_tag_score& c, counts)
      {
        if (c.tag > max_tag)             max_tag = c.tag;
        if (c.state > max_state)         max_state = c.state;
      }

    std::vector<double> tag_counts ( max_tag + 1, 0 );
    std::vector<double> state_counts ( max_state + 1, 0 );
    std::vector< std::tr1::unordered_map<size_type,double> > tag_counts_given_cluster ( max_state + 1 );
    std::vector< std::tr1::unordered_map<size_type,double> > cluster_counts_given_tag ( max_tag + 1 );
    BOOST_FOREACH(const state_tag_score& c, counts)
      {
        tag_counts [ c.tag ] += c.score;
        state_counts [ c.state ] += c.score;
        tag_counts_given_cluster [ c.state ] [ c.tag ] = c.score;
        cluster_counts_given_tag [ c.tag ] [ c.state ] = c.score;
      }
    return boost::make_tuple(tag_counts,
                             state_counts,
                             tag_counts_given_cluster,
                             cluster_counts_given_tag);
  }

  static
  double calculate_homogenity (const std::vector<double>& tag_counts,
                               const std::vector<double>& cluster_counts,
                               const std::vector< std::tr1::unordered_map<size_type,double> >& tag_counts_given_cluster)
  {
    assert ( ! cluster_counts.empty() );

    std::size_t num_clusters = cluster_counts.size();

    if (num_clusters == 1) return 1.0;
    
    const double Ht = calculate_entropy ( tag_counts.begin(), tag_counts.end() );
    const std::vector<double> probc = normalize ( cluster_counts );
    double Htc = 0.0;
    for (std::size_t c = 0; c < num_clusters; ++c)
      {
        if (probc[c] > 0)
          {
            const std::tr1::unordered_map<size_type,double>& tc = tag_counts_given_cluster [c];
            std::vector<double> counts;
            for(std::tr1::unordered_map<size_type,double>::const_iterator it = tc.begin(); 
                it != tc.end(); 
                ++it)
              {
                counts. push_back( it->second );
              }

            Htc += probc[c] * calculate_entropy ( counts.begin(), counts.end() );
          }
      }
    return 1.0 - Htc / Ht;
  }

  static
  boost::tuple<double,double,double>
  calculate_v_measure(const std::vector<state_tag_score>& counts)
  {
    const boost::tuple<std::vector<double>, 
                       std::vector<double>,
                       std::vector< std::tr1::unordered_map<size_type,double> >,
                       std::vector< std::tr1::unordered_map<size_type,double> > >
    statistics = compute_v_measure_stats ( counts );

    const std::vector<double> tag_counts = statistics.get<0>();
    const std::vector<double> cluster_counts = statistics.get<1>();
    const std::vector< std::tr1::unordered_map<size_type,double> > tag_counts_given_cluster = statistics.get<2>();
    const std::vector< std::tr1::unordered_map<size_type,double> > cluster_counts_given_tag = statistics.get<3>();
    
    const double h = calculate_homogenity (tag_counts,
                                           cluster_counts,
                                           tag_counts_given_cluster);
    const double c = calculate_homogenity (cluster_counts,
                                           tag_counts,
                                           cluster_counts_given_tag);
    const double vm = 2.0 * h * c / ( h + c);
    return boost::make_tuple ( vm, h, c );
  }

  double calculate_type_accuracy(const std::vector<size_type>& dict,
                                 const std::vector<state_tag_score>& mapping) const
  {
    assert ( ! dict.empty() );
    const size_type max_tag = *std::max_element ( dict.begin(), dict.end() );
    std::vector<size_type> state_to_tag ( max_tag + 1, 0 );
    BOOST_FOREACH(const state_tag_score& s, mapping)
      {
        state_to_tag [ s. state ] = s. tag;
      }
    std::size_t correct = 0;
    std::size_t total = 0;
    for(size_t w = 1; w < dict.size(); ++w)
      {
        const size_type t = oracle_one_tag_dict_ [w];
        if (t == 0) continue;
        const size_type p = state_to_tag [ dict [w] ];
        ++total;
        if (t == p) ++correct;
      }
    return 1.0 * correct / total;
  }

public:
  scorer(const std::vector<size_type>& wseq,
         const std::vector<size_type>& tseq)
    :
    wseq_ ( wseq ),
    tseq_ ( tseq ),
    num_tokens_ ( std::count_if(wseq.begin(), wseq.end(), boost::bind(std::not_equal_to<size_type>(), _1, 0))),
    maxt_ ( *std::max_element(tseq.begin(), tseq.end()) ),
    gold_tag_locations_ ( get_tag_locations(tseq, maxt_) ),
    oracle_one_tag_dict_ ( get_oracle_one_tag_dict(wseq, tseq) )
  {
    assert ( ! wseq.empty() );
    assert (wseq.size() == tseq.size());
  }
         
  boost::tuple<float_type,
               std::vector<size_type>,
               float_type,
               std::vector<size_type>,
               float_type,float_type,float_type,
               float_type>
  score(const std::vector<size_type>& dict) const
  {
    using namespace std;
    const vector<vector<size_type> > predicted_state_locations = get_tag_locations_from_type_dict(wseq_, dict);
    vector<state_tag_score> edge_scores = get_mapping_scores( predicted_state_locations, gold_tag_locations_ );
    sort(edge_scores.begin(), edge_scores.end(),
         boost::bind(&state_tag_score::score, _1) > boost::bind(&state_tag_score::score,_2));
    const vector<state_tag_score>& sorted_edge_scores = edge_scores;
    const size_type maxs = std::max_element(sorted_edge_scores.begin(), sorted_edge_scores.end(), state_tag_score_state_less())->state;
    const boost::tuple<float_type,vector<state_tag_score> > greedy1to1_matching = get_greedy1to1_matching(sorted_edge_scores, maxs);
    const float_type greedy1to1_accuracy = greedy1to1_matching.get<0>();
    const vector<state_tag_score>& greedy1to1_mapping = greedy1to1_matching.get<1>();
    const boost::tuple<float_type,vector<state_tag_score> > many1_matching = get_many1_matching(sorted_edge_scores, maxs);
    const float_type many1_accuracy = many1_matching.get<0>();
    const vector<state_tag_score>& many1_mapping = many1_matching.get<1>();

    const boost::tuple<double,double,double> vm_scores = calculate_v_measure ( edge_scores );
    const double vm = vm_scores.get<0>();
    const double vmh = vm_scores.get<1>();
    const double vmc = vm_scores.get<2>();

    const double type_accuracy = calculate_type_accuracy(dict, many1_mapping);

    return boost::make_tuple(greedy1to1_accuracy,
                             state_tag_score_vect_to_map(greedy1to1_mapping),
                             many1_accuracy,
                             state_tag_score_vect_to_map(many1_mapping),
                             vm,vmh,vmc,
                             type_accuracy);
  }

}; // class scorer
} // namespace

#endif // TYPETAGGER_SCORER_HPP
