#ifndef TYPETAGGER_UTILITIES_HPP
#define TYPETAGGER_UTILITIES_HPP

#include <cstdlib>
#include <fstream>
#include <vector>
#include <cassert>
#include <iostream>
#include <cmath>
#include <algorithm> // max_element
#include <utility> // pair
#include <ext/numeric> // __gnu_cxx::iota
#include <boost/scoped_ptr.hpp>
#include <boost/foreach.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/random/linear_congruential.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/uniform_real.hpp>
#include <boost/random/variate_generator.hpp>

namespace typetagger {

// Try boost::mt19937 or boost::ecuyer1988 instead of boost::minstd_rand
//typedef boost::minstd_rand boost_rng_type;
typedef boost::mt19937 boost_rng_type;
typedef boost::variate_generator<boost_rng_type&,boost::uniform_real<> > random_uniform_real_type;

std::ifstream* safe_ifopen(const char* const filename)
{
  std::ifstream *infile = new std::ifstream(filename);
  if (! *infile) 
    {
      std::cerr << "error opening file " << filename << ". abort()" << std::endl;
      abort ();
    }
  return infile;
}

std::ofstream* safe_ofopen(const char* const filename)
{
  std::ofstream *outfile = new std::ofstream(filename);
  if (! *outfile) 
    {
      std::cerr << "error opening file " << filename << ". abort()" << std::endl;
      abort ();
    }
  return outfile;
}


std::size_t count_lines(const char* const filename)
{
  using namespace std;

  boost::scoped_ptr<ifstream> infile( safe_ifopen(filename) );
  string line;
  std::size_t nsents = 0;
  while(getline(*infile, line))
    ++nsents;
  infile->close();
  return nsents;
}

std::vector<std::string> slurp(const char* const filename, const bool trim)
{
  using namespace std;

  vector<string> v;
  const size_t nsents = count_lines(filename);
  v.reserve(nsents);
  ifstream infile;
  infile.open(filename);
  string line;
  while(getline(infile, line))
    {
      if (trim) boost::algorithm::trim(line);
      v.push_back(line);
    }
  infile.close();
  assert(nsents == v.size());
  return v;
}

// int randint(const int a, const int b)
// {
//   // returns random number in closed-open interval [a,b]
//   assert (a <= b);
//   int range = b - a + 1;
//   return a + floor(1.0 * range * rand() / (1.0 + RAND_MAX));
// }

double random_uniform_real(boost_rng_type& gen)
{
  boost::uniform_real<> uni_dist(0,1);
  boost::variate_generator<boost_rng_type&,boost::uniform_real<> > uni(gen, uni_dist);
  return uni();
}

template<typename T>
std::vector<T> range(const T a, const T b)
{
  assert (a <= b);
  const std::size_t n = static_cast<std::size_t>(b-a);
  std::vector<T> v(n);
  __gnu_cxx::iota(v.begin(), v.end(), a);
  return v;
}

template<typename float_type>
boost::tuple<std::size_t,float_type,float_type>
sample_from_logs(random_uniform_real_type& rand01,
                 const std::vector<float_type>& logs)
{
  assert ( rand01.min() == 0.0 );
  assert ( rand01.max() == 1.0 );

  // calculate logz
  using namespace std;
  const float_type max_val = *( max_element(logs.begin(), logs.end()) );
  float_type z = 0.0;
  BOOST_FOREACH(const float_type val, logs)
    {
      z += exp( val - max_val );
    }
  const float_type logz = log(z);

  // calculate probs
  vector<float_type> cumulative_probs( logs.size(), 0.0 );
  float_type running_sum = 0.0;
  for(size_t i = 0; i < logs.size(); ++i)
    {
      const float_type prob_of_bin = exp(logs[i] - max_val - logz);
      running_sum += prob_of_bin;
      cumulative_probs[i] = running_sum;
    }
  if (running_sum <= 0.95 || running_sum >= 1.05)
    cout << "ERROR: typetagger::sample_from_logs.  running_sum = " << running_sum << endl;
  assert(running_sum > 0.95 && running_sum < 1.05);
  
  // sample
  //const float_type r = cumulative_probs.back() * rand() / (1.0 + RAND_MAX);
  const float_type r = cumulative_probs.back() * rand01();
  pair<typename vector<float_type>::const_iterator,typename vector<float_type>::const_iterator>
    search_result = std::equal_range(cumulative_probs.begin(), cumulative_probs.end(), r);
  const size_t index = search_result.first - cumulative_probs.begin();
  const float_type bin_prob = cumulative_probs[index] - ( index > 0 ? cumulative_probs[index-1] : 0.0 );

  return boost::make_tuple(index, bin_prob, logz);
}

template<typename InputIterator>
double calculate_entropy(InputIterator first, InputIterator last)
{
  double total = 0;
  for(InputIterator it = first; it != last; ++it)
    {
      assert (*it >= 0);
      total += *it;
    }
  double entropy = 0;
  for(InputIterator it = first; it != last; ++it)
    {
      if (*it > 0)
        {
          const double p = 1.0 * (*it) / total;
          entropy -= p * std::log(p);
        }
    }
  return entropy;
}

template<typename T>
std::vector<double> normalize(const std::vector<T>& v)
{
  const double total = std::accumulate(v.begin(), v.end(), 0.0);
  std::vector<double> w ( v.size(), 0 );
  for(std::size_t i = 0; i < v.size(); ++i)
    w [i] = 1.0 * v[i] / total;
  return w;
}

} // namespace

#endif // TYPETAGGER_UTILITIES_HPP
